try:
    from .dc_imports import discord
except ImportError:
    raise ModuleNotFoundError("No discord library found. Please install a supported library.")

commands = __import__(f"{discord.lib}.ext.commands", fromlist=[""])
tasks = __import__(f"{discord.lib}.ext.tasks", fromlist=[""])


try:
    from discord import CogMeta
    from discord.ext import bridge

    slash_command = discord.slash_command
    checks = commands

    discord.lib = "pycord"  # type: ignore

except ImportError:
    CogMeta = commands.CogMeta
    bridge = commands

    try:
        # Discord.py
        slash_command = discord.app_commands.command
        checks = discord.app_commands.checks

    except AttributeError:
        if discord.__title__ == "nextcord":
            slash_command = discord.slash_command
            checks = commands

        elif discord.__title__ == "disnake":
            slash_command = commands.slash_command
            checks = commands

        discord.lib = discord.__title__  # type: ignore


PYCORD = discord.lib == "pycord"
DPY = discord.lib == "discord"



from enum import Enum


class LogFormat(str, Enum):
    """Presets for logging formats that can be used in :func:`.set_log`.

    ``{color}`` and ``{end}`` are used to add the default color of the current log level.
    Specific colors like ``{red}`` and ``{green}`` can also be used.

    If no colors are used, the whole log message will be colored.

    ``//`` at the end of the log format is used to indicate that new lines will be fully colored.
    """

    full_color = "[%(levelname)s] %(message)s"
    full_color_time = "[%(asctime)s] %(levelname)s: %(message)s"
    color_level = "[{color}%(levelname)s{end}] %(message)s"
    color_level_time = "[{color}%(asctime)s{end}] [{color}%(levelname)s{end}] %(message)s"
    colorful = "{magenta}[%(asctime)s]{end} - {cyan}%(levelname)s{end} | %(message)s"
    default = "[{color}%(levelname)s{end}] %(message)s//"

    def __str__(self):
        return self.value


class TimeFormat(str, Enum):
    """Presets for the time format that is used in :func:`.set_log`."""

    datetime = "%Y-%m-%d %H:%M:%S"
    time = "%H:%M:%S"
    default = "%H:%M:%S"

    def __str__(self):
        return self.value


class HelpStyle(Enum):
    """Presets for the help command used in :func:`.add_help_command`."""

    embed_fields = 0
    embed_description = 1
    markdown = 2
    codeblocks = 3
    codeblocks_inline = 4

    default = embed_fields

    def __str__(self):
        return self.name


class CogLog(str, Enum):
    """Presets for log messages in :meth:`.Bot.load_cogs`.

    - ``{cog}`` is the name of the cog.
    - ``{directory}`` is the directory where a cog was loaded from.
    - ``{path}`` is the path of a cog.
    - ``{sum}`` is the amount of cogs that were loaded.

    If ``{sum}`` is used, no other variables except ``{directory}`` can be used.
    """

    sum = "Loaded {sum} cogs"
    directory_sum = "Loaded {sum} cogs from {directory}"
    path = "Loaded {path}"
    all_cogs = "Loaded {cog}"
    default = sum

    def __str__(self):
        return self.name


class ReadyEvent(Enum):
    """Styles for the ready event. This can be used in :class:`.Bot`."""

    box = 0
    box_colorful = 1
    box_bold = 2
    logs = 3
    table = 4
    table_bold = 5
    table_vertical = 6
    table_vertical_bold = 7
    default = table

    def __str__(self):
        return self.name



from __future__ import annotations

import re
from datetime import datetime, timedelta, timezone
from typing import Literal

from .errors import ConvertTimeError
from .internal import tp
from .internal.dc import discord


def set_utc(dt: datetime) -> datetime:
    """Set the timezone of a datetime object to UTC.

    Parameters
    ----------
    dt:
        The datetime object to set the timezone of.

    Returns
    -------
    :class:`datetime.datetime`
    """
    return dt.replace(tzinfo=timezone.utc)


def convert_time(seconds: int | float, relative: bool = True) -> str:
    """Convert seconds to a human-readable time.

    Parameters
    ----------
    seconds:
        The amount of seconds to convert.
    relative:
        Whether to use relative time. Defaults to ``True``.

        .. hint::
            This is only needed for German translation and will
            not have any effect if the language is set to English.

            >>> convert_time(450000, relative=True)  # Relative: Seit 5 Tagen
            '5 Tagen'
            >>> convert_time(450000, relative=False)  # Not relative: 5 Tage
            '5 Tage'

    Returns
    -------
    :class:`str`
        A human-readable time.
    """
    if seconds < 60:
        return f"{round(seconds)} {tp('sec', round(seconds))}"
    minutes = seconds / 60
    if minutes < 60:
        return f"{round(minutes)} {tp('min', round(minutes))}"
    hours = minutes / 60
    if hours < 24:
        return f"{round(hours)} {tp('hour', round(hours))}"
    days = hours / 24
    return f"{round(days)} {tp('day', round(days), relative=relative)}"


def convert_dt(dt: datetime | timedelta, relative: bool = True) -> str:
    """Convert :class:`datetime` or :class:`timedelta` to a human-readable time.

    This function calls :func:`convert_time`.

    Parameters
    ----------
    dt:
        The datetime or timedelta object to convert.
    relative:
        Whether to use relative time. Defaults to ``True``.

    Returns
    -------
    :class:`str`
        A human-readable time.
    """
    if isinstance(dt, timedelta):
        return convert_time(abs(dt.total_seconds()), relative)

    if isinstance(dt, datetime):
        return convert_time(abs((dt - discord.utils.utcnow()).total_seconds()), relative)


def dc_timestamp(
    seconds: int | float, style: Literal["t", "T", "d", "D", "f", "F", "R"] = "R"
) -> str:
    """Convert seconds to a Discord timestamp.

    Parameters
    ----------
    seconds:
        The amount of seconds to convert.
    style: :class:`str`
        The style of the timestamp. Defaults to ``R``.
        For more information, see :func:`discord.utils.format_dt`.

    Returns
    -------
    :class:`str`
        A Discord timestamp.
    """
    dt = discord.utils.utcnow() + timedelta(seconds=seconds)
    return discord.utils.format_dt(dt, style)


def convert_to_seconds(
    string: str, error: bool = False, default_unit: Literal["s", "m", "h", "d"] | None = "m"
) -> int:
    """Convert a string to seconds. Supports multiple units and decimal separators.

    Parameters
    ----------
    string:
        The string to convert.
    error:
        Whether to raise an error if the string could not be converted. If set to ``False``,
        the function will return ``0`` instead. Defaults to ``False``.
    default_unit:
        The default unit to use if no valid unit is specified. Defaults to ``m``.
        If at least one valid unit is found, all numbers without a valid unit are ignored.

    Returns
    -------
    :class:`int`
        The amount of seconds.

    Raises
    ------
    :exc:`ConvertTimeError`
         No valid number was found, or ``default_unit`` is ``None`` while no valid unit was found.

    Example
    -------
    >>> convert_to_seconds("1m 9s")
    69
    >>> convert_to_seconds("1.5m")
    90
    >>> convert_to_seconds("1,5 min")
    90
    >>> convert_to_seconds("1h 5m 10s")
    3910
    """
    units = {"s": "seconds", "m": "minutes", "h": "hours", "d": "days", "t": "days", "w": "weeks"}

    pattern = re.compile(r"(?P<value>\d+([.,]\d+)?) *(?P<unit>[smhdtw]?)", flags=re.IGNORECASE)
    matches = pattern.finditer(string)

    no_unit = "0"
    found_units = {}
    for match in matches:
        unit_char = match.group("unit").lower()
        value = float(match.group("value").replace(",", "."))

        unit = units.get(unit_char, no_unit)
        found_units[unit] = value

    if no_unit in found_units:  # Number without valid unit found
        if len(found_units) <= 1:
            # No valid unit found -> Default unit is associated with the number
            if default_unit is not None:
                found_units[units[default_unit]] = found_units[no_unit]

        del found_units[no_unit]

    if error and not found_units:
        raise ConvertTimeError(f"Could not convert '{string}' to seconds.")

    return int(timedelta(**found_units).total_seconds())



from __future__ import annotations

import asyncio
import logging
import os
import traceback
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Callable

import aiohttp
from dotenv import load_dotenv

from .emb import EzContext
from .emb import error as error_emb
from .enums import CogLog, HelpStyle, ReadyEvent
from .errors import ErrorMessageSent
from .internal import (
    READY_TITLE,
    EzConfig,
    get_error_text,
    load_lang,
    print_custom_ready,
    print_ready,
    set_lang,
    t,
)
from .internal.config import Blacklist
from .internal.dc import DPY, PYCORD, CogMeta, bridge, checks, commands, discord
from .logs import DEFAULT_LOG, custom_log, set_log
from .sql import DBHandler
from .times import dc_timestamp

if PYCORD:
    _main_bot = discord.Bot
else:
    _main_bot = commands.Bot


class Bot(_main_bot):  # type: ignore
    """The EzCord bot class. This is a subclass of :class:`discord.Bot` if you use Pycord.

    .. hint::

        As this class extends from :class:`discord.Bot`, only slash commands are supported.
        If you want to use prefix commands with ``Pycord``, use :class:`PrefixBot` instead.

        If you use ``Discord.py``, you can use this class for both slash and prefix commands.

    Parameters
    ----------
    intents:
        The intents to use for the bot. Defaults to :meth:`discord.Intents.default()`.
    debug:
        Enable log messages. Defaults to ``True``.
    error_handler:
        Enable the error handler. Defaults to ``True``.
    error_webhook_url:
        The webhook URL to send error messages to. Defaults to ``None``.

        .. note::
            You can disable the default error handler, but still provide an error webhook URL.
            This will send an error report to the webhook, but it won't send an error message to the user.
    ignored_errors:
        A list of error types to ignore. Defaults to ``None``.
    full_error_traceback:
        Whether to send the full error traceback. If this is ``False``,
        only the most recent traceback will be sent. Defaults to ``False``.
    language:
        The language to use for user output.

        The default languages are  If you add your own language file as
        described in :doc:`the language example </examples/languages>`, you can use that language as well.
    ready_event:
        The style for :meth:`on_ready_event`. Defaults to :attr:`.ReadyEvent.default`.
        If this is ``None``, the event will be disabled.
    **kwargs:
        Additional keyword arguments for :class:`discord.Bot`.
    """

    def __init__(
        self,
        intents: discord.Intents = discord.Intents.default(),
        *,
        debug: bool = True,
        error_handler: bool = True,
        error_webhook_url: str | None = None,
        ignored_errors: list[Any] | None = None,
        full_error_traceback: bool = False,
        language: str = "en",
        ready_event: ReadyEvent | None = ReadyEvent.default,
        **kwargs,
    ):
        if PYCORD:
            super().__init__(intents=intents, **kwargs)
        else:
            prefix = kwargs.pop("command_prefix", None)
            super().__init__(command_prefix=prefix or "!", intents=intents, **kwargs)

        if error_webhook_url:
            os.environ.setdefault("ERROR_WEBHOOK_URL", error_webhook_url)

        if debug:
            self.logger = set_log(DEFAULT_LOG)
        else:
            self.logger = logging.getLogger(DEFAULT_LOG)
            self.logger.addHandler(logging.NullHandler())

        self.error_handler = error_handler
        self.error_webhook_url = error_webhook_url
        self.ignored_errors = ignored_errors or []
        self.full_error_traceback = full_error_traceback
        load_lang(language)
        set_lang(language) if language != {} else set_lang("en")

        self.error_event_added = False
        if error_handler or error_webhook_url:
            self.error_event_added = True
            if DPY:
                self.tree.on_error = self._error_event
            else:
                self.add_listener(self._error_event, "on_application_command_error")

        self.ready_event = ready_event
        if ready_event:
            self.add_listener(self._ready_event, "on_ready")
        self.add_listener(self._check_cog_groups, "on_ready")
        self.add_listener(self._db_setup, "on_ready")

        self.ready_event_adds: dict = {}
        self.ready_event_removes: list[int | str] = []

        self.enabled_extensions: list[str] = []
        self.initial_cogs: list[str] = []

        # Needed for Discord.py command mentions
        self.all_dpy_commands = None

    async def get_application_context(self, interaction: discord.Interaction, cls=EzContext):
        """A custom application command context for Pycord."""
        return await super().get_application_context(interaction, cls=cls)

    def _send_cog_log(
        self,
        custom_log_level: str | None,
        log_format: str,
        color: str | None,
    ):
        if custom_log_level:
            custom_log(custom_log_level, log_format, color=color, level=logging.INFO)
        else:
            self.logger.info(log_format)

    def _cog_log(
        self,
        cog_name: str,
        custom_log_level: str | None,
        log_format: CogLog | str | None,
        directory: str,
        color: str | None = None,
    ):
        """Sends a log message for a loaded cog."""

        if not log_format or "{sum}" in log_format:
            return

        log_format = log_format.replace("{cog}", cog_name)

        dot = "." if directory else ""
        log_format = log_format.replace("{path}", f"{directory}{dot}{cog_name}")
        log_format = log_format.replace("{directory}", f"{directory}")

        self._send_cog_log(custom_log_level, log_format, color=color)

    def _cog_count_log(
        self,
        custom_log_level: str | None,
        log_format: CogLog | str | None,
        count: int,
        color: str | None = None,
        directory: str | None = None,
    ):
        """Sends a log message for the number of loaded cogs in a directory or in total."""

        if not log_format or "{cog}" in log_format or "{path}" in log_format:
            return
        if directory and "{directory}" not in log_format:
            return
        if count == 0:
            return
        if count == 1:
            log_format = log_format.replace("cogs", "cog")

        if directory and "{sum}" in log_format and "{directory}" in log_format:
            log_format = log_format.replace("{sum}", str(count))
            log_format = log_format.replace("{directory}", directory)
        elif not directory and "{sum}" in log_format and "{directory}" not in log_format:
            log_format = log_format.replace("{sum}", str(count))
        else:
            return

        self._send_cog_log(custom_log_level, log_format, color=color)

    def _manage_cogs(
        self,
        *directories: str,
        subdirectories: bool = False,
        ignored_cogs: list[str] | None,
        log: CogLog | str | None = CogLog.default,
        custom_log_level: str | None,
        log_color: str | None,
    ):
        cogs = []

        ignored_cogs = ignored_cogs or []
        if not directories:
            directories = ("cogs",)

        loaded_cogs = 0
        for directory in directories:
            for root, dirs, files in os.walk(directory):
                path = Path(root)
                loaded_dir_cogs = 0
                for filename in files:
                    name = filename[:-3]
                    if (
                        filename.endswith(".py")
                        and not filename.startswith("_")
                        and name not in ignored_cogs
                    ):
                        cogs.append(f"{'.'.join(path.parts)}.{name}")
                        loaded_dir_cogs += 1
                        self._cog_log(
                            f"{name}", custom_log_level, log, ".".join(path.parts[1:]), log_color
                        )
                loaded_cogs += loaded_dir_cogs

                self._cog_count_log(custom_log_level, log, loaded_dir_cogs, log_color, path.stem)
                if not subdirectories:
                    break
        self._cog_count_log(custom_log_level, log, loaded_cogs, log_color)
        return cogs

    def load_cogs(
        self,
        *directories: str,
        subdirectories: bool = False,
        ignored_cogs: list[str] | None = None,
        log: CogLog | str | None = CogLog.default,
        custom_log_level: str | None = "COG",
        log_color: str | None = None,
    ):
        """Load all cogs in the given directories.

        Parameters
        ----------
        *directories:
            Names of the directories to load cogs from.
            Defaults to ``"cogs"``.
        subdirectories:
            Whether to load cogs from subdirectories.
            Defaults to ``False``.
        ignored_cogs:
            A list of cogs to ignore. Defaults to ``None``.
        log:
            The log format for cogs. Defaults to :attr:`.CogLog.default`.
            If this is ``None``, logs will be disabled.
        custom_log_level:
            The name of the custom log level for cogs. Defaults to ``COG``.
        log_color:
            The color to use for cog logs. This will only have an effect if ``custom_log_level`` is enabled.
            If this is ``None``, a default color will be used.
        """

        cogs = self._manage_cogs(
            *directories,
            subdirectories=subdirectories,
            ignored_cogs=ignored_cogs,
            log=log,
            custom_log_level=custom_log_level,
            log_color=log_color,
        )
        self.initial_cogs = cogs

        if not DPY:
            for cog in cogs:
                self.load_extension(cog)

    def add_ready_info(
        self,
        name: str,
        value: str | int,
        position: int | None = None,
        color: str | None = None,
    ):
        """Adds an information to the ``on_ready`` message.

        Parameters
        ----------
        name:
            The name of the info to add. If this name already exists, the info will be updated.
        value:
            The value of the info.
        position:
            The position of the info. If this is ``None``, the info will be added at the end.
        color:
            The color of the info. If this is ``None``, a default color will be used.
        """
        self.ready_event_adds[name] = {"value": value, "position": position, "color": color}

    def remove_ready_info(self, *elements: str | int):
        """Removes an information from the ``on_ready`` message.

        Parameters
        ----------
        *elements:
            The names or positions of the infos to remove.
        """
        for element in elements:
            self.ready_event_removes.append(element)

    def ready(
        self,
        *,
        title: str = READY_TITLE,
        style: ReadyEvent = ReadyEvent.default,
        default_info: bool = True,
        new_info: dict | None = None,
        colors: list[str] | None = None,
    ):
        """Print a custom ready message.

        Parameters
        ----------
        title:
            The title of the ready message.
        style:
            The style of the ready message. Defaults to :attr:`.ReadyEvent.default`.
        default_info:
            Whether to include the default information. Defaults to ``True``.
        new_info:
            A dictionary of additional information to include in the ready message.
            Defaults to ``None``.

            .. note::
                Information can also be added with :meth:`.add_ready_info` and removed with
                :meth:`.remove_ready_info`.
        colors:
            A list of colors to use for the ready message. If no colors are given,
            default colors will be used.

            Colors can only be used with :attr:`.ReadyEvent.box_colorful` and all table styles.
        """
        modifications = self.ready_event_adds, self.ready_event_removes
        print_custom_ready(self, title, modifications, style, default_info, new_info, colors)

    async def _ready_event(self):
        """Prints the bot's information when it's ready."""
        await asyncio.sleep(0.1)

        modifications = self.ready_event_adds, self.ready_event_removes
        print_ready(self, self.ready_event, modifications=modifications)

        if DPY:
            self.all_dpy_commands = await self.tree.fetch_commands()

    @staticmethod
    async def _db_setup():
        """Calls the setup method of all registered :class:`.DBHandler` instances."""

        setup_copy = DBHandler._auto_setup.copy()
        for instance in setup_copy:
            if hasattr(instance, "setup") and callable(instance.setup):
                await instance.setup()

    async def _check_cog_groups(self):
        """Checks if all cog groups are valid."""
        for cog in self.cogs.values():
            if hasattr(cog, "group") and cog.group:
                if cog.group not in self.cogs.keys():
                    self.logger.warning(
                        f"The cog group '{cog.group}' for cog '{cog.qualified_name}' does not exist."
                    )

    async def on_error(self, event_method: str, *args: Any, **kwargs: Any) -> None:
        """This overrides the default ``on_error`` event to send an error webhook."""

        if self.error_webhook_url:
            description = f"- **Event:** {event_method}\n```py\n{traceback.format_exc()}```"
            webhook_sent = await self._send_error_webhook(description[:3750])
        else:
            webhook_sent = False

        self.logger.exception(
            f"Error in event **{event_method}** ```{traceback.format_exc(limit=0)}```",
            extra={"webhook_sent": webhook_sent},
        )

    async def _error_event(self, ctx: discord.Interaction, error: discord.DiscordException):
        """The event that handles application command errors."""
        if type(error) in self.ignored_errors + [ErrorMessageSent]:
            return

        if (
            (PYCORD and isinstance(error, discord.CheckFailure))
            or (DPY and isinstance(error, discord.app_commands.CheckFailure))
            or isinstance(error, commands.CheckFailure)
        ):
            if self.error_handler:
                await error_emb(ctx, t("no_user_perms"))
            return

        if isinstance(error, commands.CommandOnCooldown):
            if self.error_handler:
                seconds = round(ctx.command.get_cooldown_retry_after(ctx))
                cooldown_txt = t("cooldown", dc_timestamp(seconds))
                await error_emb(ctx, cooldown_txt, title=t("cooldown_title"))

        elif isinstance(error, checks.BotMissingPermissions):
            if self.error_handler:
                perms = "\n".join(error.missing_permissions)
                perm_txt = f"{t('no_perms')} ```\n{perms}```"
                await error_emb(ctx, perm_txt, title=t("no_perms_title"))

        else:
            if "original" in error.__dict__ and not self.full_error_traceback:
                original_error = error.__dict__["original"]
                error_msg = f"{original_error.__class__.__name__}: {error.__cause__}"
                error = original_error
            else:
                error_msg = f"{error}"

            if self.error_handler:
                error_txt = f"{t('error', f'```{error_msg}```')}"
                try:
                    await error_emb(ctx, error_txt, title=t("error_title"))
                except discord.HTTPException as e:
                    # ignore invalid interaction error, probably took too long to respond
                    if e.code != 10062:
                        self.logger.error("Could not send error message to user", exc_info=e)

            webhook_sent = False
            if self.error_webhook_url:
                description = get_error_text(ctx, error)
                webhook_sent = await self._send_error_webhook(description)

            self.logger.exception(
                f"Error while executing **/{ctx.command.qualified_name}** ```{error_msg}```",
                exc_info=error,
                extra={"webhook_sent": webhook_sent},
            )

    async def _send_error_webhook(self, description):
        webhook_sent = False
        async with aiohttp.ClientSession() as session:
            webhook = discord.Webhook.from_url(
                self.error_webhook_url, session=session, bot_token=self.http.token
            )

            embed = discord.Embed(
                title="Error Report",
                description=description,
                color=discord.Color.red(),
            )
            try:
                await webhook.send(
                    embed=embed,
                    username=f"{self.user.name} Error Report",
                    avatar_url=self.user.display_avatar.url,
                )
            except discord.HTTPException:
                self.logger.error(
                    "Error while sending error report to webhook. "
                    "Please check if the URL is correct."
                )
            else:
                webhook_sent = True

        return webhook_sent

    def get_cmd(self, name: str, bold: bool = True) -> str:
        """Helper method to get a command mention. Returns a string if the command was not found.

        Parameters
        ----------
        name:
            The name of the command to get.
        bold:
            Whether to bold the command name. Defaults to ``True``.
        """
        default = f"**/{name}**" if bold else f"/{name}"

        if not DPY:
            cmd = self.get_application_command(name)
            if cmd is None:
                return default
            return cmd.mention

        else:
            if not self.all_dpy_commands:
                return default

            for c in self.all_dpy_commands:
                if c.name == name:
                    return c.mention

            return default

    def add_help_command(
        self,
        *,
        style: HelpStyle = HelpStyle.default,
        embed: discord.Embed | None = None,
        show_categories: bool = True,
        show_description: bool = True,
        show_cmd_count: bool = False,
        timeout: int | None = None,
        ephemeral: bool = True,
        author_only: bool = True,
        guild_only: bool = True,
        buttons: list[discord.Button] | None = None,
        title_format: str = "{emoji} - {name}",
        description_format: str = "{description}",
        permission_check: bool = True,
    ):
        """Add a help command that uses a select menu to group commands by cogs.

        If you use :class:`Cog`, you can pass in emojis to use for the select menu.

        .. hint::
            Title and description can be formatted with the following placeholders:

            - ``{emoji}`` - The emoji of the cog.
            - ``{name}`` - The name of the cog.
            - ``{description}`` - The description of the cog.

        Parameters
        ----------
        style:
            The style to use for the help command. Defaults to :attr:`.HelpStyle.default`.
        embed:
            The embed to use for the help command. If this is ``None``, a default
            embed will be used.
        show_categories:
            Whether to display the categories of the help command front page. Defaults to ``True``.
        show_description:
            Whether to display the description in each category page. Defaults to ``True``.
        show_cmd_count:
            Whether to display the command count of each category in the select menu.
            Defaults to ``False``.
        timeout:
            The timeout for the select menu. Defaults to ``None``.
        ephemeral:
            Whether the help command should be ephemeral. Defaults to ``True``.
        author_only:
            Whether the help command should only be visible to the author. Defaults to ``True``.
            This only works if ``ephemeral`` is ``False``.
        guild_only:
            Whether the help command should only be visible in guilds. Defaults to ``True``.
        buttons:
            A list of buttons to add to the help command. Defaults to ``None``.
        title_format:
            The title format of each category.
        description_format:
            The description format of each category.
        permission_check:
            Whether to check for permissions before showing a command. Defaults to ``True``.
        """

        if buttons is None:
            buttons = []
        for button in buttons:
            if not isinstance(button, discord.ui.Button):
                raise TypeError(f"Button must be of type 'Button', not {type(button)}.")

        self.help = _CustomHelp(
            style,
            embed,
            show_categories,
            show_description,
            show_cmd_count,
            timeout,
            ephemeral,
            author_only,
            guild_only,
            buttons,
            title_format,
            description_format,
            permission_check,
        )
        self.enabled_extensions.append("help")
        if not DPY:
            self.load_extension("ezcord.cogs.pyc.help_setup", package="ezcord")

    def add_status_changer(
        self,
        activities: list[
            str | discord.Activity | discord.CustomActivity | discord.Game | discord.Streaming
        ],
        *,
        interval: int = 60,
        status: discord.Status = discord.Status.online,
        shuffle: bool = False,
        **kwargs: Callable | str,
    ):
        """Add a status changer that changes the bot's activity every ``interval`` seconds.

        .. note::

            You can use the following variables in status texts:

            - ``{guild_count}`` - The number of guilds the bot is in.
            - ``{user_count}`` - The number of users the bot can see.

            You can create custom variables by passing in variable names and values
            as ``**kwargs``.

        Parameters
        ----------
        activities:
            A list of activities to use for the status. Strings will be converted
            to :class:`discord.CustomActivity`.
        interval:
            The interval in seconds to change the status. Defaults to ``60``.
        status:
            The status to use. Defaults to :attr:`discord.Status.online`.
        shuffle:
            Whether to use a random order for the activities. Defaults to ``False``.
        **kwargs:
            Additional variables to use in status texts. This can either be a string value or
            a callable that returns a string value.

        Example
        -------
        .. code-block:: python3

            def get_coins():  # This can also be async
                return 69

            bot.add_status_changer(
                [
                    "{guild_count} Servers",  # Strings will be converted to CustomActivity
                    discord.Game("with {coins} coins")
                ],
                coins=get_coins
            )
        """

        self.status_changer = _StatusChanger(
            activities,
            interval,
            status,
            shuffle,
            kwargs,
        )
        self.enabled_extensions.append("status_changer")
        if not DPY:
            self.load_extension("ezcord.cogs.pyc.status_changer_setup", package="ezcord")

    def add_blacklist(
        self,
        admin_server_ids: list[int],
        *,
        db_path: str = "blacklist.db",
        db_name: str = "blacklist",
        raise_error: bool = False,
        owner_only: bool = True,
    ):
        """Add a blacklist that bans users from using the bot. This should be called
        before the ``on_ready`` event.

        Parameters
        ----------
        admin_server_ids:
            A list of server IDs. Admins on these servers will be able to see the admin commands.
        db_path:
            The path to the database file.
        db_name:
            The name of the database.
        raise_error:
            Whether to raise :class:`.errors.Blacklisted` error in case a blacklisted user uses
            the bot. If this is ``False``, a default message will be sent to the user.

            .. note::

                This can be used to handle the error in your own error handler.

        owner_only:
            Whether the blacklist can only be managed by the bot owner. Defaults to ``True``.
        """

        EzConfig.blacklist = Blacklist(
            db_path,
            db_name,
            raise_error,
            owner_only,
        )
        EzConfig.admin_guilds = admin_server_ids

        self.enabled_extensions.append("blacklist")
        if not DPY:
            self.load_extension("ezcord.cogs.pyc.blacklist_setup", package="ezcord")

    async def setup_hook(self):
        """This is used for Discord.py startup and should not be called manually."""

        for cog in self.initial_cogs:
            await self.load_extension(cog)

        for ext in self.enabled_extensions:
            await self.load_extension(f".cogs.dpy.{ext}_setup", package="ezcord")

    def _run_setup(
        self,
        env_path: str | os.PathLike[str] | None,
        token_var: str,
        token: str | None,
    ):
        if not env_path:
            return token

        load_dotenv(env_path)
        env_token = os.getenv(token_var)
        if token is None and env_token is not None:
            token = env_token

        if not self.error_webhook_url:
            error_webhook_url = os.environ.get("ERROR_WEBHOOK_URL")
            if error_webhook_url is not None:
                self.error_webhook_url = error_webhook_url
                if not self.error_event_added:
                    if DPY:
                        self.tree.on_error = self._error_event
                    else:
                        self.add_listener(self._error_event, "on_application_command_error")

        return token

    def run(
        self,
        token: str | None = None,
        *,
        env_path: str | os.PathLike[str] | None = ".env",
        token_var: str = "TOKEN",
        **kwargs,
    ) -> None:
        """This overrides the default :meth:`discord.Bot.run` method and automatically loads the token
        from the environment.

        Parameters
        ----------
        token:
            The bot token. If this is ``None``, the token will be loaded from the environment.
        env_path:
            The path to the environment file. Defaults to ``.env``. If this is ``None``, environment
            variables are not loaded automatically.
        token_var:
            The name of the token variable in the environment file. Defaults to ``TOKEN``.
        **kwargs:
            Additional keyword arguments for :meth:`discord.Bot.run`.
        """
        token = self._run_setup(env_path, token_var, token)
        super().run(token, **kwargs)

    async def start(
        self,
        token: str | None = None,
        *,
        env_path: str | os.PathLike[str] | None = ".env",
        token_var: str = "TOKEN",
        **kwargs,
    ) -> None:
        """This overrides the default :meth:`discord.Bot.start` method and automatically loads the token
        from the environment.

        Parameters
        ----------
        token:
            The bot token. If this is ``None``, the token will be loaded from the environment.
        env_path:
            The path to the environment file. Defaults to ``.env``. If this is ``None``, environment
            variables are not loaded automatically.
        token_var:
            The name of the token variable in the environment file. Defaults to ``TOKEN``.
        **kwargs:
            Additional keyword arguments for :meth:`discord.Bot.start`.
        """
        token = self._run_setup(env_path, token_var, token)
        await super().start(token, **kwargs)


class PrefixBot(Bot, commands.Bot):
    """A subclass of :class:`discord.ext.commands.Bot` that implements the :class:`Bot` class.

    This class can be used if you want to use EzCord with prefix commands.
    This is only needed for Pycord.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class BridgeBot(Bot, bridge.Bot):
    """A subclass of :class:`discord.ext.bridge.Bot` that implements the :class:`Bot` class.

    This class can be used if you want to use EzCord with bridge commands.
    This is only needed for Pycord.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class _CogMeta(CogMeta):
    """A metaclass for cogs that adds an ``emoji`` attribute."""

    def __new__(cls, *args, **kwargs) -> CogMeta:
        name, bases, attrs = args
        attrs["emoji"] = kwargs.pop("emoji", None)
        attrs["group"] = kwargs.pop("group", None)
        attrs["hidden"] = kwargs.pop("hidden", False)
        return super().__new__(cls, *args, **kwargs)


class Cog(commands.Cog, metaclass=_CogMeta):
    """This can be used as a base class for all cogs.

    Parameters
    ----------
    bot:
        The bot instance.
    """

    def __init__(self, bot: Bot) -> None:
        self.bot = bot


@dataclass
class _CustomHelp:
    style: HelpStyle
    embed: discord.Embed | None
    show_categories: bool
    show_description: bool
    show_cmd_count: bool
    timeout: int | None
    ephemeral: bool
    author_only: bool
    guild_only: bool
    buttons: list[discord.Button]
    title: str
    description: str
    permission_check: bool


@dataclass
class _StatusChanger:
    activities: list
    interval: int
    status: discord.Status
    shuffle: bool
    kwargs: dict[str, Callable | str]



"""Classes that are adding some functionality for default components.

.. note::

        If you want to register global checks and error handlers, all Views must
        inherit from :class:`~EzView` instead of :class:`discord.ui.View`.

        .. code-block:: python

            from ezcord import View

            class MyView(View):  # right
                ...

            class MyView(discord.ui.View):  # wrong
                ...
"""
from __future__ import annotations

import asyncio
import inspect
import os
from typing import Callable

import aiohttp

from .internal import get_error_text
from .internal.dc import PYCORD, discord
from .logs import log

_view_error_handlers: list[Callable] = []
_view_checks: list[Callable] = []
_view_check_failures: list[Callable] = []
_modal_error_handlers: list[Callable] = []


def _check_coro(func):
    if not asyncio.iscoroutinefunction(func):
        raise TypeError(f"Event registered must be a coroutine function, not {type(func)}")


def _check_params(func, amount):
    parameters = inspect.signature(func).parameters
    if len(parameters) != amount:
        raise ValueError(
            f"Event method must have '{amount}' parameters, has '{len(parameters)}' instead"
        )


def event(coro):
    """A decorator to register custom checks and error handlers for Discord components.

    Example
    -------
    .. code-block:: python

        @ezcord.event
        async def view_check(interaction):
            return interaction.user.id == 12345  # Returns True or False

        @ezcord.event
        async def on_view_check_failure(interaction):
            await interaction.response.send_message("You can't use this!")

        @ezcord.event
        async def on_view_error(error, item, interaction):
            await interaction.response.send_message("Something went wrong!")

        @ezcord.event
        async def on_modal_error(error, interaction):
            await interaction.response.send_message("Something went wrong!")
    """
    _check_coro(coro)

    name = coro.__name__.lstrip("_")

    if name == "view_check":
        _check_params(coro, 1)
        _view_checks.append(coro)
    elif name == "on_view_check_failure":
        _check_params(coro, 1)
        _view_check_failures.append(coro)
    elif name == "on_view_error":
        _check_params(coro, 3)
        _view_error_handlers.append(coro)
    elif name == "on_modal_error":
        _check_params(coro, 2)
        _modal_error_handlers.append(coro)
    else:
        raise ValueError(f"Invalid event name: '{coro.__name__}'")

    log.debug(f"Registered event **{coro.__name__}**")


async def _send_error_webhook(interaction, description) -> bool:
    error_webhook_url = os.getenv("ERROR_WEBHOOK_URL")
    if error_webhook_url is None:
        return False

    webhook_sent = False
    async with aiohttp.ClientSession() as session:
        webhook = discord.Webhook.from_url(error_webhook_url, session=session)

        embed = discord.Embed(
            title="Error Report",
            description=description,
            color=discord.Color.red(),
        )
        try:
            await webhook.send(
                embed=embed,
                username=f"{interaction.client.user.name} Error Report",
                avatar_url=interaction.client.user.display_avatar.url,
            )
        except discord.HTTPException:
            log.error(
                "Error while sending error report to webhook. "
                "Please check if the URL is correct."
            )
        else:
            webhook_sent = True

    return webhook_sent


class EzView(discord.ui.View):
    """This class extends from :class:`discord.ui.View` and adds some functionality."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    async def on_error(
        self, error: Exception, item: discord.ui.Item, interaction: discord.Interaction
    ) -> None:
        """Sends an error message to a webhook, if the URL was passed in :class:`.Bot`.

        Executes all registered error handlers with the ``@ezcord.event`` decorator.
        """
        if not PYCORD:
            error, item, interaction = item, interaction, error

        description = get_error_text(interaction, error, item)
        webhook_sent = await _send_error_webhook(interaction, description)

        log.exception(
            f"Error in View **{type(item.view).__name__}** ```{error}```",
            exc_info=error,
            extra={"webhook_sent": webhook_sent},
        )

        for error_coro in _view_error_handlers:
            await error_coro(error, item, interaction)

    async def interaction_check(self, interaction: discord.Interaction) -> bool:
        """Returns ``True`` if all custom checks return ``True`` or if no custom checks are registered."""
        for coro in _view_checks:
            if not await coro(interaction):
                return False
        return True

    async def on_check_failure(self, interaction: discord.Interaction) -> None:
        """This method is called if :meth:`interaction_check` returns ``False``."""
        for coro in _view_check_failures:
            await coro(interaction)

    async def on_timeout(self) -> None:
        """If ``disable_on_timeout`` is set to ``True``, this will disable all components,
        unless the viw has been explicitly stopped.
        """
        if not PYCORD:
            return await super().on_timeout()

        if self.disable_on_timeout:
            self.disable_all_items()

            # Fixes Pycord's NotFound error if message is ephemeral
            # and interaction has already been responded to
            try:
                message = self.parent or self._message  # type: ignore
            except AttributeError:
                # Older Pycord versions are missing parent attribute
                return await super().on_timeout()

            if message:
                try:
                    m = await message.edit(view=self)
                except discord.NotFound:
                    return
                if m:
                    self._message = m


class EzModal(discord.ui.Modal):
    """This class extends from :class:`discord.ui.Modal` and adds an error handler."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    async def on_error(self, error: Exception, interaction: discord.Interaction) -> None:
        """Sends an error message to a webhook, if the webhook URL was passed into :class:`.Bot`.

        Executes all registered error handlers with the ``@ezcord.event`` decorator.
        """
        if not PYCORD:
            error, interaction = interaction, error

        description = get_error_text(interaction, error, self)
        webhook_sent = await _send_error_webhook(interaction, description)

        log.exception(
            f"Error in Modal **{type(self).__name__}**",
            exc_info=error,
            extra={"webhook_sent": webhook_sent},
        )

        for error_coro in _modal_error_handlers:
            await error_coro(error, interaction)


Modal = EzModal
View = EzView



from datetime import datetime, timedelta

import pytest

import ezcord
from ezcord import ConvertTimeError


def test_convert_time():
    dt = ezcord.set_utc(datetime.utcnow())

    assert isinstance(ezcord.convert_dt(dt), str)
    assert isinstance(ezcord.convert_dt(timedelta(seconds=5)), str)
    assert isinstance(ezcord.convert_time(5), str)


def test_dc_timestamp():
    result = ezcord.dc_timestamp(0)
    assert result.startswith("<t:") and result.endswith(":R>")


def test_convert_so_seconds():
    assert ezcord.convert_to_seconds("1m 9s") == 69
    assert ezcord.convert_to_seconds("1.5m") == 90
    assert ezcord.convert_to_seconds("1,5 min") == 90
    assert ezcord.convert_to_seconds("1h 5m 10s") == 3910

    # tests with no units
    assert ezcord.convert_to_seconds("1 2m 3") == 120
    assert ezcord.convert_to_seconds("2") == 120
    assert ezcord.convert_to_seconds("2", default_unit="s", error=True) == 2
    assert ezcord.convert_to_seconds("2", default_unit=None) == 0

    with pytest.raises(ConvertTimeError):
        ezcord.convert_to_seconds("1 2 3", default_unit=None, error=True)

    # tests with invalid units
    assert ezcord.convert_to_seconds("") == 0
    assert ezcord.convert_to_seconds("z") == 0

    with pytest.raises(ConvertTimeError):
        assert ezcord.convert_to_seconds("", error=True)

    with pytest.raises(ConvertTimeError):
        assert ezcord.convert_to_seconds("z", error=True)



from .internal.dc import DPY, discord

if DPY:
    _DC_ERROR_TYPE = discord.app_commands.AppCommandError
else:
    _DC_ERROR_TYPE = discord.DiscordException


class ErrorMessageSent(_DC_ERROR_TYPE):  # type: ignore
    """Exception that can be raised to indicate that an error message has already been sent to the user.

    This could be useful if an error message has already been sent to the user within a check
    function. This prevents further messages to the user, as this error will be ignored by
    the default error handler.
    """


class EzcordException(Exception):
    """Base exception class for all Ezcord exceptions."""


class ConvertTimeError(EzcordException):
    """Raised when a time conversion fails."""


class Blacklisted(EzcordException):
    """Can be raised when a blacklisted user tries to use a command.

    This error can be caught in a command error handler to send a custom response.
    """



from .colors import *
from .config import EzConfig
from .embed_templates import *
from .funcutils import copy_kwargs
from .language.languages import load_lang
from .ready_style import *
from .translation import *



from __future__ import annotations

import inspect
import random
from copy import deepcopy

from .. import emb
from ..bot import Bot, Cog
from ..components import View
from ..enums import HelpStyle
from ..internal import replace_embed_values, t
from ..internal.dc import PYCORD, discord, slash_command
from ..logs import log


def get_emoji(cog: Cog) -> str:
    if hasattr(cog, "emoji") and cog.emoji:
        emoji = cog.emoji
    else:
        emoji = random.choice(["🔰", "👻", "🍪", "👥", "🦕", "🐧", "✨", "😩", "🍔"])

    return emoji


def get_group(cog: Cog) -> str | None:
    if hasattr(cog, "group") and cog.group:
        return cog.group
    return None


def replace_placeholders(s: str, **kwargs: str):
    for key, value in kwargs.items():
        if not value:
            continue
        s = s.replace(f"{{{key}}}", value)
    return s


def get_perm_parent(cmd: discord.SlashCommand) -> discord.SlashCommandGroup | None:
    """Iterates through parent groups until it finds a group with default_member_permissions set."""
    if PYCORD:
        while cmd.default_member_permissions is None:
            cmd = cmd.parent
            if cmd is None:
                return None
    else:
        while cmd.default_permissions is None:
            cmd = cmd.parent
            if cmd is None:
                return None

    return cmd


async def pass_checks(command: discord.SlashCommand, ctx) -> bool:
    """Returns True if the current user passes all checks."""
    passed = True
    for check in deepcopy(command.checks):
        try:
            if inspect.iscoroutinefunction(check):
                await check(ctx)
            else:
                if not check(ctx):
                    passed = False
                    break
        except Exception:
            passed = False
            break

    return passed


class Help(Cog, hidden=True):
    def __init__(self, bot: Bot):
        super().__init__(bot)
        self.help.guild_only = bot.help.guild_only

    @slash_command(name=t("cmd_name"), description=t("cmd_description"))
    async def help(self, ctx):
        embed = self.bot.help.embed
        if embed is None:
            embed = discord.Embed(title=t("embed_title"), color=discord.Color.blue())
        else:
            interaction = ctx.interaction if PYCORD else ctx
            embed = replace_embed_values(embed, interaction)

        options = []
        commands: dict[str, dict] = {}
        for name, cog in self.bot.cogs.items():
            if hasattr(cog, "hidden") and cog.hidden:
                continue

            group = get_group(cog)
            name = group if group else name
            name = name.title()

            if len(name) == 0:
                log.warning(
                    "A cog has a name with length 0. "
                    "This cog will not be displayed in the help command."
                )
                continue

            if len(name) > 100:
                name = name[:90] + "..."

            if name not in commands:
                commands[name] = {}
            if "cmds" not in commands[name]:
                commands[name]["cmds"] = []

            if len(cog.get_commands()) == 0 and PYCORD:
                continue

            emoji = get_emoji(cog)
            commands[name]["emoji"] = emoji

            desc = cog.description
            if not cog.description:
                desc = t("default_description", name)
                if not desc:
                    log.warning(
                        f"The default description for cog '{name}' is invalid. "
                        f"This can be changed in the language file."
                    )

            commands[name]["description"] = desc

            field_name = replace_placeholders(self.bot.help.title, name=name, emoji=emoji)
            desc = replace_placeholders(
                self.bot.help.description, description=desc, name=name, emoji=emoji
            )

            if PYCORD:
                cog_cmds = [
                    cmd
                    for cmd in cog.walk_commands()
                    if type(cmd)
                    not in [discord.MessageCommand, discord.UserCommand, discord.SlashCommandGroup]
                ]
            else:
                cog_cmds = cog.walk_app_commands()

            for command in cog_cmds:
                if PYCORD:
                    default_perms = command.default_member_permissions
                    guild_ids = command.guild_ids
                else:
                    default_perms = command.default_permissions
                    guild_ids = command._guild_ids

                if self.bot.help.permission_check:
                    if not await pass_checks(command, ctx):
                        continue

                    if default_perms and not command.parent:
                        if not default_perms.is_subset(ctx.user.guild_permissions):
                            continue

                    parent = get_perm_parent(command)
                    if parent and not parent.default_member_permissions.is_subset(
                        ctx.user.guild_permissions
                    ):
                        continue

                if ctx.guild and guild_ids and ctx.guild.id not in guild_ids:
                    continue

                commands[name]["cmds"].append(command)

            cmd_count = len(commands[name]["cmds"])

            if cmd_count == 0:
                continue
            if not group:
                if self.bot.help.show_cmd_count:
                    label = f"{name} ({cmd_count})"
                else:
                    label = name
                option = discord.SelectOption(label=label, emoji=emoji, value=name)
                options.append(option)
                if self.bot.help.show_categories:
                    embed.add_field(name=field_name, value=desc, inline=False)

        if len(options) == 0:
            return await ctx.response.send_message(t("no_commands"), ephemeral=True)
        if len(options) > 25 or len(embed.fields) > 25:
            log.error(
                f"Help command category limit reached. Only 25 out of {len(options)} are shown."
            )
            options = options[:25]
            embed.fields = embed.fields[:25]
        view = CategoryView(options, self.bot, ctx.user, commands)
        for button in self.bot.help.buttons:
            view.add_item(deepcopy(button))
        await ctx.response.send_message(view=view, embed=embed, ephemeral=self.bot.help.ephemeral)


class CategorySelect(discord.ui.Select):
    def __init__(
        self,
        options: list[discord.SelectOption],
        bot: Bot,
        member: discord.Member | discord.User,
        commands: dict[str, dict],
    ):
        super().__init__(min_values=1, max_values=1, placeholder=t("placeholder"), options=options)
        self.bot = bot
        self.member = member
        self.commands = commands

    def get_mention(self, cmd) -> str:
        """This is only needed for Discord.py."""
        if self.bot.all_dpy_commands:
            for c in self.bot.all_dpy_commands:
                if c.name == cmd.name:
                    cmd = c
                    break

        try:
            return cmd.mention
        except AttributeError:
            return f"**/{cmd.name}**"

    async def callback(self, interaction: discord.Interaction):
        if self.bot.help.author_only and interaction.user != self.member:
            return await emb.error(interaction, t("wrong_user"))

        cmds = self.commands[self.values[0]]
        title = self.values[0].title()
        emoji = cmds["emoji"]

        embed = self.bot.help.embed
        if embed is None:
            embed = discord.Embed(
                color=discord.Color.blue(),
            )
        else:
            embed = replace_embed_values(embed, interaction)
        embed.title = replace_placeholders(self.bot.help.title, name=title, emoji=emoji)
        embed.clear_fields()

        if self.bot.help.show_description:
            embed.description = desc = cmds["description"] + "\n"
        else:
            desc = ""

        commands = cmds["cmds"]
        embed_field_styles = [
            HelpStyle.embed_fields,
            HelpStyle.codeblocks_inline,
            HelpStyle.codeblocks,
        ]
        style = self.bot.help.style
        if len(commands) > 25 and style in embed_field_styles:
            style = HelpStyle.embed_description

        if style == HelpStyle.embed_fields:
            for command in commands:
                embed.add_field(
                    name=f"**{self.get_mention(command)}**",
                    value=f"`{command.description}`",
                    inline=False,
                )

        elif style == HelpStyle.codeblocks or style == HelpStyle.codeblocks_inline:
            for command in commands:
                embed.add_field(
                    name=f"**{self.get_mention(command)}**",
                    value=f"```{command.description}```",
                    inline=style == HelpStyle.codeblocks_inline,
                )

        elif style == HelpStyle.embed_description:
            embed.description = desc + "\n"
            for command in commands:
                if len(embed.description) <= 3500:
                    embed.description += (
                        f"**{self.get_mention(command)}**\n{command.description}\n\n"
                    )
                else:
                    log.error("Help embed length limit reached. Some commands are not shown.")
                    break

        elif style == HelpStyle.markdown:
            embed.description = desc
            for command in commands:
                if len(embed.description) <= 3500:
                    embed.description += f"### {self.get_mention(command)}\n{command.description}\n"
                else:
                    log.error("Help embed length limit reached. Some commands are not shown.")
                    break

        if len(commands) == 0:
            embed.description = t("no_commands")

        view = CategoryView(self.options, self.bot, self.member, self.commands)
        for button in self.bot.help.buttons:
            view.add_item(deepcopy(button))
        await interaction.response.edit_message(embed=embed, view=view)


class CategoryView(View):
    def __init__(
        self,
        options: list[discord.SelectOption],
        bot: Bot,
        member: discord.Member | discord.User,
        commands: dict[str, dict],
    ):
        if PYCORD:
            super().__init__(timeout=bot.help.timeout, disable_on_timeout=True)
        else:
            super().__init__(timeout=None)

        self.add_item(CategorySelect(options, bot, member, commands))



import inspect
import random
from itertools import cycle

from ..bot import Bot, Cog
from ..internal.dc import discord, tasks


class Activity(Cog, hidden=True):
    def __init__(self, bot: Bot):
        super().__init__(bot)

        activities = [
            discord.CustomActivity(name=act) if isinstance(act, str) else act
            for act in self.bot.status_changer.activities
        ]

        if self.bot.status_changer.shuffle:
            random.shuffle(activities)

        self.activities = cycle(activities)

    @discord.ext.commands.Cog.listener()
    async def on_ready(self):
        if not self.change_activity.is_running():
            self.change_activity.change_interval(seconds=self.bot.status_changer.interval)
            self.change_activity.start()

    @tasks.loop()
    async def change_activity(self):
        """Replaces default variables and user variables in the activity name."""
        replace_values = {
            "guild_count": f"{len(self.bot.guilds):,}",
            "user_count": f"{len(self.bot.users):,}",
        }

        act = next(self.activities)

        for var, replace_value in replace_values.items():
            act.name = act.name.replace("{" + var + "}", str(replace_value))

        for key, value in self.bot.status_changer.kwargs.items():
            if inspect.iscoroutinefunction(value):
                replace_value = await value()
            elif callable(value):
                replace_value = value()
            else:
                replace_value = value

            act.name = act.name.replace("{" + str(key) + "}", str(replace_value))

        # Not sure why this is needed, but it is.
        if act.type == discord.ActivityType.custom:
            act = discord.CustomActivity(name=act.name)

        await self.bot.change_presence(activity=act, status=self.bot.status_changer.status)



from __future__ import annotations

import traceback
from copy import deepcopy
from functools import cache

from ..internal.dc import discord
from .config import EzConfig

_TEMPLATES: dict[str, discord.Embed] = {
    "success_embed": discord.Embed(color=discord.Color.green()),
    "error_embed": discord.Embed(color=discord.Color.red()),
    "warn_embed": discord.Embed(color=discord.Color.gold()),
    "info_embed": discord.Embed(color=discord.Color.blue()),
}


def save_embeds(**kwargs: discord.Embed | str):
    """Save embeds to config.

    If one of the default values is not included, a default template will be saved.
    """
    embeds = {}
    overrides = _TEMPLATES if len(kwargs) == 0 else kwargs

    for name, embed in overrides.items():
        if embed is None:
            embeds[name] = _TEMPLATES[name].to_dict()
        elif isinstance(embed, str):
            embeds[name] = embed
        else:
            embeds[name] = embed.to_dict()

    EzConfig.embed_templates = embeds


@cache
def load_embed(name: str) -> discord.Embed | str:
    """Load an embed template."""

    if not EzConfig.embed_templates:
        save_embeds()

    embeds = EzConfig.embed_templates

    if isinstance(embeds[name], str):
        return embeds[name]

    try:
        return discord.Embed.from_dict(embeds[name])
    except KeyError:
        if name in _TEMPLATES.keys():
            save_embeds()
            return load_embed()
        else:
            raise ValueError(f"Embed template '{name}' not found.")


def format_error(error: Exception) -> str:
    txt = "".join(traceback.format_exception(type(error), error, error.__traceback__))
    return f"\n```py\n{txt[:3500]}```"


def get_error_text(
    ctx,
    error: Exception,
    item: discord.ui.Item | discord.ui.Modal | None = None,
):
    """Get the description for the webhook embed."""
    if item:
        if isinstance(item, discord.ui.Button) and item.label:
            location = f"- **Button:** {item.label}"
        elif ctx.type == discord.InteractionType.modal_submit:
            location = f"- **Modal:** {type(item).__name__}"
        else:
            location = f"- **Select Menu:** {type(item.view).__name__}"
    else:
        location = f"- **Command:** /{ctx.command.qualified_name}"

    guild_txt = f"\n- **Guild:** {ctx.guild.name} - `{ctx.guild.id}`" if ctx.guild else ""
    user_txt = f"\n- **User:** {ctx.user} - `{ctx.user.id}`" if ctx.user else ""

    description = location + guild_txt + user_txt + format_error(error)
    return description


def replace_values(s: str, interaction: discord.Interaction) -> str:
    user = interaction.user
    s = s.replace("{user}", f"{user}")
    s = s.replace("{username}", user.name)
    s = s.replace("{user_mention}", user.mention)
    s = s.replace("{user_id}", f"{user.id}")
    s = s.replace("{user_avatar}", user.display_avatar.url)

    if interaction.guild:
        s = s.replace("{servername}", interaction.guild.name)
    else:
        s = s.replace("{servername}", interaction.client.user.name)

    if interaction.guild and interaction.guild.icon:
        s = s.replace("{server_icon}", interaction.guild.icon.url)
    else:
        s = s.replace("{server_icon}", interaction.client.user.display_avatar.url)

    return s


def replace_dict(content: dict | str, interaction: discord.Interaction) -> dict | str:
    """Recursively loop through a dictionary and replace certain values
    with information from the current interaction.
    """
    if isinstance(content, str):
        return replace_values(content, interaction)

    for key, value in content.items():
        if isinstance(value, str):
            content[key] = replace_values(value, interaction)
        elif isinstance(value, list):
            items = []
            for element in value:
                items.append(replace_dict(element, interaction))
            content[key] = items
        elif isinstance(value, dict):
            content[key] = replace_dict(value, interaction)

    return content


def replace_embed_values(embed: discord.Embed, interaction: discord.Interaction):
    embed_dict = deepcopy(embed).to_dict()
    embed_dict = replace_dict(embed_dict, interaction)
    return discord.Embed.from_dict(embed_dict)



"""
The blacklist can be managed within Discord using the ``/admin`` commands.
However, if you want to manage the blacklist in your own code, you can use these functions.

.. warning::

    These methods can only be used after the blacklist has been activated with :meth:`.add_blacklist`.
"""

from __future__ import annotations

from .internal import EzConfig
from .sql import DBHandler


class _BanDB(DBHandler):
    def __init__(self):
        self.db_name = EzConfig.blacklist.db_name
        super().__init__(EzConfig.blacklist.db_path)

    async def setup(self):
        await self.exec(
            f"""CREATE TABLE IF NOT EXISTS {self.db_name} (
            user_id INTEGER PRIMARY KEY,
            reason TEXT,
            dt DATETIME DEFAULT CURRENT_TIMESTAMP
            )"""
        )

    async def add_ban(self, user_id: int, reason: str | None):
        await self.exec(
            f"INSERT INTO {self.db_name} (user_id, reason) VALUES (?, ?)", (user_id, reason)
        )

    async def remove_ban(self, user_id: int):
        result = await self.exec(f"DELETE FROM {self.db_name} WHERE user_id = ?", (user_id,))
        return result.rowcount

    async def get_bans(self):
        return await self.all(f"SELECT user_id FROM {self.db_name}")

    async def get_full_bans(self):
        return await self.all(f"SELECT user_id, reason, dt FROM {self.db_name}")


def _blacklist_ready() -> bool:
    if not EzConfig.blacklist:
        return False
    return True


async def get_bans() -> list[int]:
    """Get all banned user IDs."""
    if not _blacklist_ready():
        return []
    return await _BanDB().get_bans()


async def get_full_bans() -> list[tuple[int, str | None, str]]:
    """Get all information about banned users.

    Returns
    -------
        A list of tuples containing the user ID, ban reason, and date of the ban.
    """
    if not _blacklist_ready():
        return []
    return await _BanDB().get_full_bans()


async def add_ban(user_id: int, reason: str | None = None) -> bool:
    """Ban a user.

    Parameters
    ----------
    user_id:
        The user ID to ban
    reason:
        The reason for the ban

    Returns
    -------
    bool:
        True if the user was banned successfully, False otherwise
    """
    if not _blacklist_ready():
        return False
    await _BanDB().add_ban(user_id, reason)
    return True


async def remove_ban(user_id: int) -> bool:
    """Unban a user.

    Parameters
    ----------
    user_id:
        The user ID to unban

    Returns
    -------
    bool:
        True if the user was unbanned successfully, False otherwise
    """
    if not _blacklist_ready():
        return False
    result = await _BanDB().remove_ban(user_id)
    return bool(result)



import sys
from typing import Any, Callable, TypeVar, cast

if sys.version_info >= (3, 10):
    from typing import ParamSpec
else:
    from typing_extensions import ParamSpec

P = ParamSpec("P")
T = TypeVar("T")


def copy_kwargs(kwargs_call: Callable[P, Any]) -> Callable[[Callable[..., T]], Callable[P, T]]:
    """Decorator returning a cast of the original function."""

    def return_func(func: Callable[..., T]) -> Callable[P, T]:
        return cast(Callable[P, T], func)

    return return_func
